"use strict";
var __classPrivateFieldSet = (this && this.__classPrivateFieldSet) || function (receiver, state, value, kind, f) {
    if (kind === "m") throw new TypeError("Private method is not writable");
    if (kind === "a" && !f) throw new TypeError("Private accessor was defined without a setter");
    if (typeof state === "function" ? receiver !== state || !f : !state.has(receiver)) throw new TypeError("Cannot write private member to an object whose class did not declare it");
    return (kind === "a" ? f.call(receiver, value) : f ? f.value = value : state.set(receiver, value)), value;
};
var __classPrivateFieldGet = (this && this.__classPrivateFieldGet) || function (receiver, state, kind, f) {
    if (kind === "a" && !f) throw new TypeError("Private accessor was defined without a getter");
    if (typeof state === "function" ? receiver !== state || !f : !state.has(receiver)) throw new TypeError("Cannot read private member from an object whose class did not declare it");
    return kind === "m" ? f : kind === "a" ? f.call(receiver) : f ? f.value : state.get(receiver);
};
var _StageRange_from, _StageRange_to;
Object.defineProperty(exports, "__esModule", { value: true });
exports.AsyncSeriesHook = exports.AsyncParallelHook = exports.SyncHook = exports.StageRange = void 0;
class Hook {
    constructor(args, name) {
        this.args = args;
        this.name = name;
        this.taps = [];
        this.interceptors = [];
    }
    intercept(interceptor) {
        this.interceptors.push(Object.assign({}, interceptor));
        if (interceptor.register) {
            for (let i = 0; i < this.taps.length; i++) {
                this.taps[i] = interceptor.register(this.taps[i]);
            }
        }
    }
    _runRegisterInterceptors(options) {
        for (const interceptor of this.interceptors) {
            if (interceptor.register) {
                const newOptions = interceptor.register(options);
                if (newOptions !== undefined) {
                    options = newOptions;
                }
            }
        }
        return options;
    }
    _runCallInterceptors(...args) {
        for (const interceptor of this.interceptors) {
            if (interceptor.call) {
                interceptor.call(...args);
            }
        }
    }
    _runErrorInterceptors(e) {
        for (const interceptor of this.interceptors) {
            if (interceptor.error) {
                interceptor.error(e);
            }
        }
    }
    _runTapInterceptors(tap) {
        for (const interceptor of this.interceptors) {
            if (interceptor.tap) {
                interceptor.tap(tap);
            }
        }
    }
    _runDoneInterceptors() {
        for (const interceptor of this.interceptors) {
            if (interceptor.done) {
                interceptor.done();
            }
        }
    }
    isUsed() {
        return this.taps.length > 0 || this.interceptors.length > 0;
    }
    callAsyncStageRange(stageRange, ...args) {
        throw new Error("Hook should implement there own _callAsyncStageRange");
    }
    callAsync(...args) {
        return this.callAsyncStageRange(StageRange.all(), ...args);
    }
    promiseStageRange(stageRange, ...args) {
        return new Promise((resolve, reject) => {
            // @ts-expect-error
            this.callAsyncStageRange(stageRange, ...args, (e, r) => {
                if (e)
                    return reject(e);
                return resolve(r);
            });
        });
    }
    promise(...args) {
        return this.promiseStageRange(StageRange.all(), ...args);
    }
    tap(options, fn) {
        this._tap("sync", options, fn);
    }
    _tap(type, options, fn) {
        if (typeof options === "string") {
            options = {
                name: options.trim()
            };
        }
        else if (typeof options !== "object" || options === null) {
            throw new Error("Invalid tap options");
        }
        if (typeof options.name !== "string" || options.name === "") {
            throw new Error("Missing name for tap");
        }
        let insert = Object.assign({ type, fn }, options);
        insert = this._runRegisterInterceptors(insert);
        if (insert.stage) {
            insert.stage = StageRange.trim(insert.stage);
        }
        this._insert(insert);
    }
    _insert(item) {
        let before;
        if (typeof item.before === "string") {
            before = new Set([item.before]);
        }
        else if (Array.isArray(item.before)) {
            before = new Set(item.before);
        }
        let stage = 0;
        if (typeof item.stage === "number") {
            stage = item.stage;
        }
        let i = this.taps.length;
        while (i > 0) {
            i--;
            const x = this.taps[i];
            this.taps[i + 1] = x;
            const xStage = x.stage || 0;
            if (before) {
                if (before.has(x.name)) {
                    before.delete(x.name);
                    continue;
                }
                if (before.size > 0) {
                    continue;
                }
            }
            if (xStage > stage) {
                continue;
            }
            i++;
            break;
        }
        this.taps[i] = item;
    }
}
class StageRange {
    constructor(from, to) {
        _StageRange_from.set(this, void 0);
        _StageRange_to.set(this, void 0);
        __classPrivateFieldSet(this, _StageRange_from, StageRange.trim(from), "f");
        __classPrivateFieldSet(this, _StageRange_to, StageRange.trim(to), "f");
    }
    static from(from, to) {
        return new StageRange(from, to);
    }
    static all() {
        return StageRange.from(StageRange.MIN, StageRange.MAX);
    }
    get from() {
        return __classPrivateFieldGet(this, _StageRange_from, "f");
    }
    get to() {
        return __classPrivateFieldGet(this, _StageRange_to, "f");
    }
    static trim(n) {
        if (n > StageRange.MAX)
            return StageRange.MAX;
        if (n < StageRange.MIN)
            return StageRange.MIN;
        return n;
    }
}
_StageRange_from = new WeakMap(), _StageRange_to = new WeakMap();
StageRange.MAX = 2 ** 31 - 1;
StageRange.MIN = -(2 ** 31);
exports.StageRange = StageRange;
class SyncHook extends Hook {
    callAsyncStageRange({ from, to }, ...args) {
        var _a;
        const args2 = [...args];
        const cb = args2.pop();
        if (from === StageRange.MIN) {
            this._runCallInterceptors(...args2);
        }
        for (let tap of this.taps) {
            const stage = (_a = tap.stage) !== null && _a !== void 0 ? _a : 0;
            if (from < stage && stage <= to) {
                this._runTapInterceptors(tap);
                try {
                    tap.fn(...args2);
                }
                catch (e) {
                    const err = e;
                    this._runErrorInterceptors(err);
                    return cb(err);
                }
            }
        }
        if (to === StageRange.MAX) {
            this._runDoneInterceptors();
            cb(null);
        }
    }
    call(...args) {
        return this.callStageRange(StageRange.all(), ...args);
    }
    /**
     * call a range of taps, from < stage <= to, (from, to]
     */
    callStageRange(stageRange, ...args) {
        let result, error;
        // @ts-expect-error
        this.callAsyncStageRange(stageRange, ...args, (e, r) => {
            error = e;
            result = r;
        });
        if (error) {
            throw error;
        }
        return result;
    }
    tapAsync() {
        throw new Error("tapAsync is not supported on a SyncHook");
    }
    tapPromise() {
        throw new Error("tapPromise is not supported on a SyncHook");
    }
}
exports.SyncHook = SyncHook;
class AsyncParallelHook extends Hook {
    callAsyncStageRange({ from, to }, ...args) {
        var _a;
        const args2 = [...args];
        const cb = args2.pop();
        if (from === StageRange.MIN) {
            this._runCallInterceptors(...args2);
        }
        const done = () => {
            this._runDoneInterceptors();
            cb(null);
        };
        const error = (e) => {
            this._runErrorInterceptors(e);
            cb(e);
        };
        const tapsInRange = [];
        for (let tap of this.taps) {
            const stage = (_a = tap.stage) !== null && _a !== void 0 ? _a : 0;
            if (from < stage && stage <= to) {
                tapsInRange.push(tap);
            }
        }
        if (tapsInRange.length === 0)
            return done();
        let counter = tapsInRange.length;
        for (let tap of tapsInRange) {
            this._runTapInterceptors(tap);
            if (tap.type === "promise") {
                const promise = tap.fn(...args2);
                if (!promise || !promise.then) {
                    throw new Error("Tap function (tapPromise) did not return promise (returned " +
                        promise +
                        ")");
                }
                promise.then(() => {
                    counter -= 1;
                    if (counter === 0) {
                        done();
                    }
                }, (e) => {
                    counter = 0;
                    error(e);
                });
            }
            else if (tap.type === "async") {
                tap.fn(...args2, (e) => {
                    if (e) {
                        counter = 0;
                        error(e);
                    }
                    else {
                        counter -= 1;
                        if (counter === 0) {
                            done();
                        }
                    }
                });
            }
            else {
                let hasError = false;
                try {
                    tap.fn(...args2);
                }
                catch (e) {
                    hasError = true;
                    counter = 0;
                    error(e);
                }
                if (!hasError && --counter === 0) {
                    done();
                }
            }
            if (counter <= 0)
                return;
        }
    }
    tapAsync(options, fn) {
        this._tap("async", options, fn);
    }
    tapPromise(options, fn) {
        this._tap("promise", options, fn);
    }
}
exports.AsyncParallelHook = AsyncParallelHook;
class AsyncSeriesHook extends Hook {
    callAsyncStageRange({ from, to }, ...args) {
        var _a;
        const args2 = [...args];
        const cb = args2.pop();
        if (from === StageRange.MIN) {
            this._runCallInterceptors(...args2);
        }
        const done = () => {
            this._runDoneInterceptors();
            cb(null);
        };
        const error = (e) => {
            this._runErrorInterceptors(e);
            cb(e);
        };
        const tapsInRange = [];
        for (let tap of this.taps) {
            const stage = (_a = tap.stage) !== null && _a !== void 0 ? _a : 0;
            if (from < stage && stage <= to) {
                tapsInRange.push(tap);
            }
        }
        if (tapsInRange.length === 0)
            return done();
        let index = 0;
        const next = () => {
            const tap = tapsInRange[index];
            this._runTapInterceptors(tap);
            if (tap.type === "promise") {
                const promise = tap.fn(...args2);
                if (!promise || !promise.then) {
                    throw new Error("Tap function (tapPromise) did not return promise (returned " +
                        promise +
                        ")");
                }
                promise.then(() => {
                    index += 1;
                    if (index === tapsInRange.length) {
                        done();
                    }
                    else {
                        next();
                    }
                }, (e) => {
                    index = tapsInRange.length;
                    error(e);
                });
            }
            else if (tap.type === "async") {
                tap.fn(...args2, (e) => {
                    if (e) {
                        index = tapsInRange.length;
                        error(e);
                    }
                    else {
                        index += 1;
                        if (index === tapsInRange.length) {
                            done();
                        }
                        else {
                            next();
                        }
                    }
                });
            }
            else {
                let hasError = false;
                try {
                    tap.fn(...args2);
                }
                catch (e) {
                    hasError = true;
                    index = tapsInRange.length;
                    error(e);
                }
                if (!hasError) {
                    index += 1;
                    if (index === tapsInRange.length) {
                        done();
                    }
                    else {
                        next();
                    }
                }
            }
            if (index === tapsInRange.length)
                return;
        };
        next();
    }
    tapAsync(options, fn) {
        this._tap("async", options, fn);
    }
    tapPromise(options, fn) {
        this._tap("promise", options, fn);
    }
}
exports.AsyncSeriesHook = AsyncSeriesHook;
